# -*- coding: utf-8 -*-
from efc.utils import join_functions
from weakref import WeakKeyDictionary


class MetaSingleCellOperandCache(type):
    _sources = WeakKeyDictionary()

    def _bind_source(cls, source):
        def clear_cache():
            if source in cls._sources:
                cls._sources[source].clear()

        def remove_cell(ws_name, row, column):
            if source in cls._sources:
                cache = cls._sources[source]

                # TODO so ugly
                for row_fixed, column_fixed in ((False, False), (False, True), (True, False), (True, True)):
                    key = (ws_name, row, column, row_fixed, column_fixed)
                    if key in cache:
                        del cache[key]

        if getattr(source, 'clear_cache_bind', None) is not None:
            source.clear_cache_bind = join_functions(clear_cache, source.clear_cache_bind)
        else:
            source.clear_cache_bind = clear_cache

        if getattr(source, 'remove_cell_bind', None) is not None:
            source.remove_cell_bind = join_functions(remove_cell, source.remove_cell_bind)
        else:
            source.remove_cell_bind = remove_cell

    def _get_source_cache(cls, source):
        try:
            return cls._sources[source]
        except KeyError:
            cache = cls._sources[source] = {}
            cls._bind_source(source)
            return cache

    def __call__(cls, row, column, row_fixed=False, column_fixed=False, ws_name=None, source=None):
        if source is not None and ws_name is not None and source.use_cache:
            cache = cls._get_source_cache(source)
            key = (ws_name, row, column, row_fixed, column_fixed)

            try:
                return cache[key]
            except KeyError:
                value = cache[key] = super(MetaSingleCellOperandCache, cls).__call__(row, column, row_fixed=row_fixed,
                                                                                     column_fixed=column_fixed,
                                                                                     ws_name=ws_name,
                                                                                     source=source)
                return value
        else:
            return super(MetaSingleCellOperandCache, cls).__call__(row, column, row_fixed=row_fixed,
                                                                   column_fixed=column_fixed, ws_name=ws_name,
                                                                   source=source)


class MetaCellRangeOperandCache(type):
    _sources = WeakKeyDictionary()

    def _bind_source(cls, source):
        def clear_cache():
            if source in cls._sources:
                cls._sources[source].clear()

        def remove_cell(ws_name, row, column):
            if source in cls._sources:
                cache = cls._sources[source]

                # TODO so ugly
                source._remove_from_range_cache(ws_name, row, column, cache)

        if getattr(source, 'clear_cache_bind', None) is not None:
            source.clear_cache_bind = join_functions(clear_cache, source.clear_cache_bind)
        else:
            source.clear_cache_bind = clear_cache

        if getattr(source, 'remove_cell_bind', None) is not None:
            source.remove_cell_bind = join_functions(remove_cell, source.remove_cell_bind)
        else:
            source.remove_cell_bind = remove_cell

    def _get_source_cache(cls, source):
        try:
            return cls._sources[source]
        except KeyError:
            cache = cls._sources[source] = {}
            cls._bind_source(source)
            return cache

    def __call__(cls, row1, column1, row2, column2,
                 row1_fixed=False, column1_fixed=False, row2_fixed=False, column2_fixed=False,
                 ws_name=None, source=None):
        if source is not None and ws_name is not None and source.use_cache:
            cache = cls._get_source_cache(source)
            key = (ws_name, row1, column1, row2, column2, row1_fixed, column1_fixed, row2_fixed, column2_fixed)

            try:
                return cache[key]
            except KeyError:
                value = cache[key] = super(MetaCellRangeOperandCache, cls).__call__(
                    row1, column1, row2, column2, row1_fixed, column1_fixed, row2_fixed, column2_fixed, ws_name, source
                )
                return value
        else:
            return super(MetaCellRangeOperandCache, cls).__call__(row1, column1, row2, column2, row1_fixed,
                                                                  column1_fixed, row2_fixed, column2_fixed, ws_name, source)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import pytest

from efc.utils import col_str_to_index


@pytest.mark.parametrize(
    ['col', 'index'],
    (
            ('A', 1),
            ('C', 3),
            ('Z', 26),
            ('AA', 27),
            ('HPP', 5840),
    )
)
def test_col_str_to_index(col, index):
    assert col_str_to_index(col) == index



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals


class BaseExcelInterface(object):
    """
    Base class to working with excel document
    """
    clear_cache_bind = None
    remove_cell_bind = None
    use_cache = False

    def __init__(self):
        self.ifs_range_cache = {} if self.use_cache else None

    def clear_cache(self):
        if self.clear_cache_bind is not None:
            self.clear_cache_bind()

    def remove_cell_cache(self, ws_name, row, column):
        if self.remove_cell_bind is not None:
            self.remove_cell_bind(ws_name, row, column)

        if self.ifs_range_cache:
            self._remove_from_range_cache(ws_name, row, column, self.ifs_range_cache)

    @staticmethod
    def _remove_from_range_cache(ws_name, row, column, cache):
        for key in list(cache):
            if ws_name == key[0]:
                if key[1] is None:
                    if key[2] >= column >= key[4]:
                        del cache[key]
                elif key[2] is None:
                    if key[1] >= row >= key[3]:
                        del cache[key]
                elif key[1] >= row >= key[3] and key[2] >= column >= key[4]:
                    del cache[key]

    def cell_to_value(self, row, column, ws_name):
        """
        :type row: int
        :type column: int
        :type ws_name: basestring
        :rtype: list
        """
        raise NotImplementedError

    def named_range_to_cells(self, name, ws_name):
        """
        TEST_RANGE -> SingleCellOperand
        OTHER_RANGE -> CellRangeOperand
        OTHER2_RANGE -> CellSetOperand
        :type name: basestring
        :type ws_name: basestring
        :rtype: list
        """
        raise NotImplementedError

    def max_row(self, ws_name):
        raise NotImplementedError

    def min_row(self, ws_name):
        raise NotImplementedError

    def max_column(self, ws_name):
        raise NotImplementedError

    def min_column(self, ws_name):
        raise NotImplementedError

    def has_worksheet(self, ws_name):
        raise NotImplementedError

    def has_named_range(self, name, ws_name):
        raise NotImplementedError



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from efc.rpn_builder.lexer import Lexer
from efc.rpn_builder.parser import Parser
from efc.rpn_builder.rpn import RPN

__author__ = 'Gleb Orlov <orlovgb@mail.ru>'
__version__ = '0.1.30'


def calc(formula, ws_name, source):
    return Parser().to_rpn(Lexer().parse(formula), ws_name, source).calc(formula, ws_name, source)


def get_calculator():
    lexer = Lexer()
    parser = Parser()

    def calculate(formula, ws_name, source):
        tokens_line = lexer.parse(formula)
        rpn = parser.to_rpn(tokens_line, ws_name, source)
        return rpn.calc(ws_name, source)

    return calculate



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import pytest

from efc import get_calculator
from efc.rpn_builder.parser.operands import WorksheetNotExist
from .mock import ExcelMock


@pytest.fixture(scope='session')
def calc():
    source = ExcelMock()
    calculator = get_calculator()
    return lambda line, ws_name: calculator(line, ws_name, source)


def test_cell_address(calc):
    assert calc('A3', 'Sheet 1').value == 4
    assert calc('C1', 'Sheet 1').value == 18

    assert calc('B100', 'Yet another sheet').value == 2
    assert calc('AA104', 'Yet another sheet').value == 45

    with pytest.raises(WorksheetNotExist):
        calc('F104', 'Some error ws').value

    assert calc('Sheet4!A3', 'Yet another sheet').value == 4
    assert calc('\'Sheet 1\'!C1', 'Yet another sheet').value == 18

    # test arith with address
    assert calc('A3 + 3', 'Sheet 1').value == 7
    assert calc('C1 / 9', 'Sheet 1').value == 2

    assert calc('Sheet4!A3 ^ 2', 'Yet another sheet').value == 16
    assert calc('\'Sheet 1\'!C1 - 4 - 1', 'Yet another sheet').value == 13

    result = calc('Sheet4!A1:B3', 'Yet another sheet')
    assert [c.value for c in result.value] == [13, 16, 13, 16, 4, 2]

    assert calc('Sheet4!test', 'Yet another sheet').value == 16
    assert calc('SUM(Sheet4!test2)', 'Yet another sheet').value == 34
    assert calc('SUM([0]Sheet4!test2)', 'Yet another sheet').value == 34


def test_single_cell_cache():
    # Cache disabled
    source = ExcelMock()
    calculator = get_calculator()

    op1 = calculator('A3', 'Sheet 1', source)
    op2 = calculator('A3', 'Sheet 1', source)

    assert op1 is not op2

    # Cache enabled
    source.use_cache = True
    op1 = calculator('A3', 'Sheet 1', source)
    op2 = calculator('A3', 'Sheet 1', source)

    assert op1 is op2

    # New cache value for A3
    source.clear_cache()
    op2 = calculator('A3', 'Sheet 1', source)
    assert op1 is not op2

    # New cache value for A3
    source.remove_cell_cache('Sheet 1', 3, 1)
    op3 = calculator('A3', 'Sheet 1', source)
    assert op2 is not op3




# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from efc.rpn_builder.lexer.tokens import (AddToken, ArithmeticToken, CellsRangeToken, ConcatToken, DivideToken,
                                          ExponentToken, FunctionToken, LeftBracketToken, MultiplyToken,
                                          NamedRangeToken, OperandToken, OperationToken, RightBracketToken, Separator,
                                          SingleCellToken, SubtractToken)
from efc.rpn_builder.parser.errors import InconsistentParentheses, SeparatorBlockError
from efc.rpn_builder.parser.operands import (
    CellRangeOperand, EmptyOperand, NamedRangeOperand, RPNOperand, SimpleOperand, SingleCellOperand,
)
from efc.rpn_builder.parser.operations import ArithmeticOperation, FunctionOperation, Operation
from efc.rpn_builder.rpn import RPN

__all__ = ('Parser',)

OPERATORS_PRIORITY = {
    ExponentToken: 5,
    MultiplyToken: 4,
    DivideToken: 4,
    SubtractToken: 3,
    AddToken: 2,
    ConcatToken: 1,
}


class Parser(object):
    @staticmethod
    def get_priority(token):
        return OPERATORS_PRIORITY.get(token.__class__, 0)

    @staticmethod
    def operand_token_handler(token, ws_name, source):
        if isinstance(token, SingleCellToken):
            return SingleCellOperand(row=token.row, column=token.column,
                                     row_fixed=token.row_fixed, column_fixed=token.column_fixed,
                                     ws_name=token.ws_name or ws_name,
                                     source=source)
        elif isinstance(token, CellsRangeToken):
            return CellRangeOperand(row1=token.row1, column1=token.column1,
                                    row1_fixed=token.row1_fixed, column1_fixed=token.column1_fixed,
                                    row2=token.row2, column2=token.column2,
                                    row2_fixed=token.row2_fixed, column2_fixed=token.column2_fixed,
                                    ws_name=token.ws_name or ws_name,
                                    source=source)
        elif isinstance(token, NamedRangeToken):
            return NamedRangeOperand(name=token.name,
                                     ws_name=token.ws_name or ws_name,
                                     source=source).value
        else:
            return SimpleOperand(value=token.token_value,
                                 ws_name=ws_name,
                                 source=source)

    def operation_token_handler(self, line):
        current_token = line.current()
        prev_token = line.prev()
        operation = ArithmeticOperation(current_token.src_value, self.get_priority(current_token))
        if (isinstance(current_token, (SubtractToken, AddToken)) and (
                isinstance(prev_token, (ArithmeticToken, LeftBracketToken, Separator)) or prev_token is None)):
            operation.operands_count = 1
        return operation

    def handle_result(self, result, ws_name, source):
        if not result:
            return EmptyOperand(ws_name=ws_name, source=source)
        elif len(result) == 1:
            return result[0]
        else:
            return RPNOperand(result, ws_name=ws_name, source=source)

    def to_rpn(self, line, ws_name, source, is_operand=False):
        result = RPN(line.src_line)
        stack = []

        result_append = result.append
        stack_append = stack.append
        stack_pop = stack.pop
        while not line.is_ended:
            token = next(line)
            if isinstance(token, OperandToken):
                result_append(self.operand_token_handler(token, ws_name, source))
            elif isinstance(token, FunctionToken):
                stack_append(FunctionOperation(token.src_value))
            elif isinstance(token, LeftBracketToken):
                stack_append(token)
                if isinstance(line.prev(), FunctionToken):
                    result_append(self.to_rpn(line, ws_name, source, is_operand=True))
            elif isinstance(token, OperationToken):
                operation = self.operation_token_handler(line)
                while stack:
                    if isinstance(stack[-1], Operation) and stack[-1].priority >= operation.priority:
                        result_append(stack_pop())
                    else:
                        break
                stack_append(operation)
            elif isinstance(token, RightBracketToken):
                while stack:
                    top_stack_token = stack_pop()
                    if isinstance(top_stack_token, LeftBracketToken):
                        break
                    else:
                        result_append(top_stack_token)
                else:
                    if is_operand:
                        line.step_back()
                        return self.handle_result(result, ws_name, source)
                    raise InconsistentParentheses(ws_name, line.src_line)
            elif isinstance(token, Separator):
                try:
                    while not isinstance(stack[-1], LeftBracketToken):
                        result_append(stack_pop())
                except IndexError:
                    if is_operand:
                        line.step_back()
                        return self.handle_result(result, ws_name, source)
                    raise SeparatorBlockError(ws_name, line.src_line)

                if len(stack) > 1 and isinstance(stack[-2], FunctionOperation):
                    stack[-2].operands_count += 1
                result_append(self.to_rpn(line, ws_name, source, is_operand=True))

        for stack_token in reversed(stack):
            if isinstance(stack_token, LeftBracketToken):
                raise InconsistentParentheses(ws_name, line.src_line)
            result_append(stack_token)

        return result



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import pytest
from six.moves import zip

from efc.rpn_builder.lexer import Lexer, tokens


@pytest.fixture(scope='module')
def lexer():
    return Lexer()


@pytest.mark.parametrize(
    ['line', 'token_types'],
    (
            ('FALSE', [tokens.BoolToken]),
            ('TRUE', [tokens.BoolToken]),
            ('4', [tokens.IntToken]),
            ('5.54', [tokens.FloatToken]),
            ('"hello"', [tokens.StringToken]),
            ('A4', [tokens.SingleCellToken]),
            ('$A$4', [tokens.SingleCellToken]),
            ('$A4', [tokens.SingleCellToken]),
            ('A$4', [tokens.SingleCellToken]),
            ('A$4:AAA5', [tokens.CellsRangeToken]),
            ('\'List\'!$A$4', [tokens.SingleCellToken]),
            ('\'List 1\'!A$4', [tokens.SingleCellToken]),
            ('\'List 1\'!A$4:AAA5', [tokens.CellsRangeToken]),
            ('\'List 1\'!hello', [tokens.NamedRangeToken]),
            ('SUM(', [tokens.FunctionToken, tokens.LeftBracketToken]),
            ('SUM', [tokens.NamedRangeToken]),
            ('SUM TRUE', [tokens.NamedRangeToken, tokens.BoolToken]),
            ('Hello Mister', [tokens.NamedRangeToken, tokens.NamedRangeToken]),
            ('SUM(34,43)',
             [tokens.FunctionToken, tokens.LeftBracketToken, tokens.IntToken,
              tokens.Separator, tokens.IntToken, tokens.RightBracketToken]),
    )
)
def test_operands_parse(lexer, line, token_types):
    parsed_line = lexer.parse(line)
    assert len(parsed_line) == len(token_types), 'Len of tokens lines not equal'

    for c, token in zip(token_types, parsed_line):
        assert isinstance(token, c)


@pytest.mark.parametrize(
    ['line', 'token_types'],
    (
            ('+', [tokens.AddToken]),
            ('-', [tokens.SubtractToken]),
            ('/', [tokens.DivideToken]),
            ('*', [tokens.MultiplyToken]),
            ('&', [tokens.ConcatToken]),
            ('^', [tokens.ExponentToken]),
            ('<>', [tokens.CompareNotEqToken]),
            ('>=', [tokens.CompareGTEToken]),
            ('<=', [tokens.CompareLTEToken]),
            ('>', [tokens.CompareGTToken]),
            ('<', [tokens.CompareLTToken]),
            ('=', [tokens.CompareEqToken]),
            ('4 + 5.54 - "hello"',
             [tokens.IntToken, tokens.AddToken, tokens.FloatToken, tokens.SubtractToken,
              tokens.StringToken]),
            ('4+5.54-"hello"+\'List 1\'!hello',
             [tokens.IntToken, tokens.AddToken, tokens.FloatToken, tokens.SubtractToken,
              tokens.StringToken, tokens.AddToken, tokens.NamedRangeToken]),
    )
)
def test_operators_parse(lexer, line, token_types):
    parsed_line = lexer.parse(line)
    assert len(parsed_line) == len(token_types), 'Len of tokens lines not equal'

    for c, token in zip(token_types, parsed_line):
        assert isinstance(token, c)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import pytest

from efc.rpn_builder.lexer.errors import CheckSumError
from efc.utils import u


def test_check_sum_error_class():
    src_line = '1 + 2'
    parsed_line = '1+2'

    e = CheckSumError(src_line=src_line, parsed_line=parsed_line)
    template = 'Code 100. Some symbols from line are lost. Src line: {src_line}. Parsed line: {parsed_line}.'
    assert u(e) == template.format(src_line=src_line, parsed_line=parsed_line)

    with pytest.raises(CheckSumError):
        raise e
